@@ -1,5 +1,5 @@
 /* GConf
- * Copyright (C) 1999, 2000 Red Hat Inc.
+ * Copyright (C) 1999, 2000, 2002 Red Hat Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -25,13 +25,32 @@
 #include <string.h>
 #include <stdlib.h>
 
+typedef struct {
+  GConfValueType type;
+  union {
+    gchar* string_data;
+    gint int_data;
+    gboolean bool_data;
+    gdouble float_data;
+    GConfSchema* schema_data;
+    struct {
+      GConfValueType type;
+      GSList* list;
+    } list_data;
+    struct {
+      GConfValue* car;
+      GConfValue* cdr;
+    } pair_data;
+  } d;
+} GConfRealValue;
+
+#define REAL_VALUE(x) ((GConfRealValue*)(x))
+
 static void
 set_string(gchar** dest, const gchar* src)
 {
-  if (*dest != NULL)
-    g_free(*dest);
-
-  *dest = src ? g_strdup(src) : NULL;
+  g_free(*dest);
+  *dest = g_strdup(src);
 }
 
 /*
@@ -42,16 +61,23 @@ GConfValue*
 gconf_value_new(GConfValueType type)
 {
   GConfValue* value;
-
-  g_return_val_if_fail(GCONF_VALUE_TYPE_VALID(type), NULL);
+  static gboolean initted = FALSE;
   
-  /* Probably want to use mem chunks here eventually. */
-  value = g_new0(GConfValue, 1);
+  g_return_val_if_fail(GCONF_VALUE_TYPE_VALID(type), NULL);
+
+  if (!initted)
+    {
+      _gconf_init_i18n ();
+      initted = TRUE;
+    }
+  
+  value = (GConfValue*) g_slice_new0 (GConfRealValue);
 
   value->type = type;
 
   /* the g_new0() is important: sets list type to invalid, NULLs all
-     pointers */
+   * pointers
+   */
   
   return value;
 }
@@ -62,6 +88,9 @@ gconf_value_new_from_string(GConfValueType type, const gchar* value_str,
 {
   GConfValue* value;
 
+  g_return_val_if_fail (type != GCONF_VALUE_LIST, NULL);
+  g_return_val_if_fail (type != GCONF_VALUE_PAIR, NULL);
+
   value = gconf_value_new(type);
 
   switch (type)
@@ -513,7 +542,7 @@ gconf_value_new_pair_from_string(GConfValueType car_type,
 }
 
 gchar*
-gconf_value_to_string(GConfValue* value)
+gconf_value_to_string(const GConfValue* value)
 {
   /* These strings shouldn't be translated; they're primarily 
      intended for machines to read, not humans, though I do
@@ -681,28 +710,30 @@ copy_value_list(GSList* list)
 }
 
 GConfValue* 
-gconf_value_copy(GConfValue* src)
+gconf_value_copy (const GConfValue* src)
 {
-  GConfValue* dest;
-
+  GConfRealValue *dest;
+  GConfRealValue *real;
+  
   g_return_val_if_fail(src != NULL, NULL);
 
-  dest = gconf_value_new(src->type);
+  real = REAL_VALUE (src);
+  dest = REAL_VALUE (gconf_value_new (src->type));
 
-  switch (src->type)
+  switch (real->type)
     {
     case GCONF_VALUE_INT:
     case GCONF_VALUE_FLOAT:
     case GCONF_VALUE_BOOL:
     case GCONF_VALUE_INVALID:
-      dest->d = src->d;
+      dest->d = real->d;
       break;
     case GCONF_VALUE_STRING:
-      set_string(&dest->d.string_data, src->d.string_data);
+      set_string(&dest->d.string_data, real->d.string_data);
       break;
     case GCONF_VALUE_SCHEMA:
-      if (src->d.schema_data)
-        dest->d.schema_data = gconf_schema_copy(src->d.schema_data);
+      if (real->d.schema_data)
+        dest->d.schema_data = gconf_schema_copy(real->d.schema_data);
       else
         dest->d.schema_data = NULL;
       break;
@@ -711,22 +742,22 @@ gconf_value_copy(GConfValue* src)
       {
         GSList* copy;
 
-        copy = copy_value_list(src->d.list_data.list);
+        copy = copy_value_list(real->d.list_data.list);
 
         dest->d.list_data.list = copy;
-        dest->d.list_data.type = src->d.list_data.type;
+        dest->d.list_data.type = real->d.list_data.type;
       }
       break;
       
     case GCONF_VALUE_PAIR:
 
-      if (src->d.pair_data.car)
-        dest->d.pair_data.car = gconf_value_copy(src->d.pair_data.car);
+      if (real->d.pair_data.car)
+        dest->d.pair_data.car = gconf_value_copy(real->d.pair_data.car);
       else
         dest->d.pair_data.car = NULL;
 
-      if (src->d.pair_data.cdr)
-        dest->d.pair_data.cdr = gconf_value_copy(src->d.pair_data.cdr);
+      if (real->d.pair_data.cdr)
+        dest->d.pair_data.cdr = gconf_value_copy(real->d.pair_data.cdr);
       else
         dest->d.pair_data.cdr = NULL;
 
@@ -736,17 +767,21 @@ gconf_value_copy(GConfValue* src)
       g_assert_not_reached();
     }
   
-  return dest;
+  return (GConfValue*) dest;
 }
 
 static void
 gconf_value_free_list(GConfValue* value)
 {
   GSList* tmp;
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_LIST);
+
+  real = REAL_VALUE (value);
   
-  tmp = value->d.list_data.list;
+  tmp = real->d.list_data.list;
 
   while (tmp != NULL)
     {
@@ -754,41 +789,176 @@ gconf_value_free_list(GConfValue* value)
       
       tmp = g_slist_next(tmp);
     }
-  g_slist_free(value->d.list_data.list);
+  g_slist_free(real->d.list_data.list);
 
-  value->d.list_data.list = NULL;
+  real->d.list_data.list = NULL;
 }
 
 void 
 gconf_value_free(GConfValue* value)
 {
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
 
-  switch (value->type)
+  real = REAL_VALUE (value);
+  
+  switch (real->type)
     {
     case GCONF_VALUE_STRING:
-      if (value->d.string_data != NULL)
-        g_free(value->d.string_data);
+      g_free(real->d.string_data);
       break;
     case GCONF_VALUE_SCHEMA:
-      if (value->d.schema_data != NULL)
-        gconf_schema_free(value->d.schema_data);
+      if (real->d.schema_data != NULL)
+        gconf_schema_free(real->d.schema_data);
       break;
     case GCONF_VALUE_LIST:
       gconf_value_free_list(value);
       break;
     case GCONF_VALUE_PAIR:
-      if (value->d.pair_data.car != NULL)
-        gconf_value_free(value->d.pair_data.car);
+      if (real->d.pair_data.car != NULL)
+        gconf_value_free(real->d.pair_data.car);
 
-      if (value->d.pair_data.cdr != NULL)
-        gconf_value_free(value->d.pair_data.cdr);
+      if (real->d.pair_data.cdr != NULL)
+        gconf_value_free(real->d.pair_data.cdr);
       break;
     default:
       break;
     }
   
-  g_free(value);
+  g_slice_free(GConfRealValue, real);
+}
+
+const char*
+gconf_value_get_string (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_STRING, NULL);
+  
+  return REAL_VALUE (value)->d.string_data;
+}
+
+char*
+gconf_value_steal_string (GConfValue *value)
+{
+  char *string;
+  GConfRealValue *real;
+  
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_STRING, NULL);
+
+  real = REAL_VALUE (value);
+
+  string = real->d.string_data;
+  real->d.string_data = NULL;
+
+  return string;
+}
+
+int
+gconf_value_get_int (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, 0);
+  g_return_val_if_fail (value->type == GCONF_VALUE_INT, 0);
+  
+  return REAL_VALUE (value)->d.int_data;
+}
+
+double
+gconf_value_get_float (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, 0.0);
+  g_return_val_if_fail (value->type == GCONF_VALUE_FLOAT, 0.0);
+  
+  return REAL_VALUE (value)->d.float_data;
+}
+
+GConfValueType
+gconf_value_get_list_type (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, GCONF_VALUE_INVALID);
+  g_return_val_if_fail (value->type == GCONF_VALUE_LIST, GCONF_VALUE_INVALID);
+  
+  return REAL_VALUE (value)->d.list_data.type;
+}
+
+GSList*
+gconf_value_get_list (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_LIST, NULL);
+
+  return REAL_VALUE (value)->d.list_data.list;
+}
+
+GSList*
+gconf_value_steal_list (GConfValue *value)
+{
+  GSList *list;
+  GConfRealValue *real;
+  
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_LIST, NULL);
+
+  real = REAL_VALUE (value);
+
+  list = real->d.list_data.list;
+  real->d.list_data.list = NULL;
+  return list;
+}
+
+GConfValue*
+gconf_value_get_car (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_PAIR, NULL);
+  
+  return REAL_VALUE (value)->d.pair_data.car;
+}
+
+GConfValue*
+gconf_value_get_cdr (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_PAIR, NULL);
+  
+  return REAL_VALUE (value)->d.pair_data.cdr;
+}
+
+
+gboolean
+gconf_value_get_bool (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, FALSE);
+  g_return_val_if_fail (value->type == GCONF_VALUE_BOOL, FALSE);
+  
+  return REAL_VALUE (value)->d.bool_data;
+}
+
+GConfSchema*
+gconf_value_get_schema (const GConfValue *value)
+{
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_SCHEMA, NULL);
+  
+  return REAL_VALUE (value)->d.schema_data;
+}
+
+GConfSchema*
+gconf_value_steal_schema (GConfValue *value)
+{
+  GConfSchema *schema;
+  GConfRealValue *real;
+  
+  g_return_val_if_fail (value != NULL, NULL);
+  g_return_val_if_fail (value->type == GCONF_VALUE_SCHEMA, NULL);
+
+  real = REAL_VALUE (value);
+
+  schema = real->d.schema_data;
+  real->d.schema_data = NULL;
+
+  return schema;
 }
 
 void        
@@ -797,16 +967,29 @@ gconf_value_set_int(GConfValue* value, gint the_int)
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_INT);
 
-  value->d.int_data = the_int;
+  REAL_VALUE (value)->d.int_data = the_int;
 }
 
 void        
 gconf_value_set_string(GConfValue* value, const gchar* the_str)
+{  
+  gconf_value_set_string_nocopy (value,
+                                 g_strdup (the_str));
+}
+
+void
+gconf_value_set_string_nocopy (GConfValue *value,
+                               char       *str)
 {
+  GConfRealValue *real;
+
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_STRING);
 
-  set_string(&value->d.string_data, the_str);
+  real = REAL_VALUE (value);
+
+  g_free (real->d.string_data);
+  real->d.string_data = str;
 }
 
 void        
@@ -815,7 +998,7 @@ gconf_value_set_float(GConfValue* value, gdouble the_float)
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_FLOAT);
 
-  value->d.float_data = the_float;
+  REAL_VALUE (value)->d.float_data = the_float;
 }
 
 void        
@@ -824,36 +1007,44 @@ gconf_value_set_bool(GConfValue* value, gboolean the_bool)
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_BOOL);
 
-  value->d.bool_data = the_bool;
+  REAL_VALUE (value)->d.bool_data = the_bool;
 }
 
 void       
-gconf_value_set_schema(GConfValue* value, GConfSchema* sc)
+gconf_value_set_schema(GConfValue* value, const GConfSchema* sc)
 {
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_SCHEMA);
-  
-  if (value->d.schema_data != NULL)
-    gconf_schema_free(value->d.schema_data);
 
-  value->d.schema_data = gconf_schema_copy(sc);
+  real = REAL_VALUE (value);
+  
+  if (real->d.schema_data != NULL)
+    gconf_schema_free (real->d.schema_data);
+
+  real->d.schema_data = gconf_schema_copy (sc);
 }
 
 void        
 gconf_value_set_schema_nocopy(GConfValue* value, GConfSchema* sc)
 {
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_SCHEMA);
   g_return_if_fail(sc != NULL);
-  
-  if (value->d.schema_data != NULL)
-    gconf_schema_free(value->d.schema_data);
 
-  value->d.schema_data = sc;
+  real = REAL_VALUE (value);
+  
+  if (real->d.schema_data != NULL)
+    gconf_schema_free (real->d.schema_data);
+
+  real->d.schema_data = sc;
 }
 
 void
-gconf_value_set_car(GConfValue* value, GConfValue* car)
+gconf_value_set_car(GConfValue* value, const GConfValue* car)
 {
   gconf_value_set_car_nocopy(value, gconf_value_copy(car));
 }
@@ -861,17 +1052,21 @@ gconf_value_set_car(GConfValue* value, GConfValue* car)
 void
 gconf_value_set_car_nocopy(GConfValue* value, GConfValue* car)
 {
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_PAIR);
-  
-  if (value->d.pair_data.car != NULL)
-    gconf_value_free(value->d.pair_data.car);
 
-  value->d.pair_data.car = car;
+  real = REAL_VALUE (value);
+  
+  if (real->d.pair_data.car != NULL)
+    gconf_value_free (real->d.pair_data.car);
+
+  real->d.pair_data.car = car;
 }
 
 void
-gconf_value_set_cdr(GConfValue* value, GConfValue* cdr)
+gconf_value_set_cdr(GConfValue* value, const GConfValue* cdr)
 {
   gconf_value_set_cdr_nocopy(value, gconf_value_copy(cdr));
 }
@@ -879,60 +1074,284 @@ gconf_value_set_cdr(GConfValue* value, GConfValue* cdr)
 void
 gconf_value_set_cdr_nocopy(GConfValue* value, GConfValue* cdr)
 {
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_PAIR);
-  
-  if (value->d.pair_data.cdr != NULL)
-    gconf_value_free(value->d.pair_data.cdr);
 
-  value->d.pair_data.cdr = cdr;
+  real = REAL_VALUE (value);
+  
+  if (real->d.pair_data.cdr != NULL)
+    gconf_value_free (real->d.pair_data.cdr);
+
+  real->d.pair_data.cdr = cdr;
 }
 
 void
-gconf_value_set_list_type(GConfValue* value,
+gconf_value_set_list_type (GConfValue    *value,
                            GConfValueType type)
 {
+  GConfRealValue *real;
+  
   g_return_if_fail(value != NULL);
   g_return_if_fail(value->type == GCONF_VALUE_LIST);
   g_return_if_fail(type != GCONF_VALUE_LIST);
   g_return_if_fail(type != GCONF_VALUE_PAIR);
-  /* If the list is non-NULL either we already have the right
-     type, or we shouldn't be changing it without deleting
-     the list first. */
-  g_return_if_fail(value->d.list_data.list == NULL);
 
-  value->d.list_data.type = type;
+  real = REAL_VALUE (value);
+  
+  /* If the list is non-NULL either we already have the right
+   * type, or we shouldn't be changing it without deleting
+   * the list first.
+   */
+  g_return_if_fail (real->d.list_data.list == NULL);
+
+  real->d.list_data.type = type;
 }
 
 void
-gconf_value_set_list_nocopy(GConfValue* value,
+gconf_value_set_list_nocopy (GConfValue* value,
                              GSList* list)
 {
-  g_return_if_fail(value != NULL);
-  g_return_if_fail(value->type == GCONF_VALUE_LIST);
-  g_return_if_fail(value->d.list_data.type != GCONF_VALUE_INVALID);
+  GConfRealValue *real;
   
-  if (value->d.list_data.list)
-    gconf_value_free_list(value);
+  g_return_if_fail (value != NULL);
+  g_return_if_fail (value->type == GCONF_VALUE_LIST);
 
-  value->d.list_data.list = list;
+  real = REAL_VALUE (value);
+  
+  g_return_if_fail (real->d.list_data.type != GCONF_VALUE_INVALID);
+  
+  if (real->d.list_data.list)
+    gconf_value_free_list (value);
+
+  real->d.list_data.list = list;
 }
 
 void
 gconf_value_set_list       (GConfValue* value,
-                             GSList* list)
+                            GSList* list)
 {
-  g_return_if_fail(value != NULL);
-  g_return_if_fail(value->type == GCONF_VALUE_LIST);
-  g_return_if_fail(value->d.list_data.type != GCONF_VALUE_INVALID);
-  g_return_if_fail((list == NULL) ||
-                   ((list->data != NULL) &&
-                    (((GConfValue*)list->data)->type == value->d.list_data.type)));
+  GConfRealValue *real;
   
-  if (value->d.list_data.list)
-    gconf_value_free_list(value);
+  g_return_if_fail (value != NULL);
+  g_return_if_fail (value->type == GCONF_VALUE_LIST);
 
-  value->d.list_data.list = copy_value_list(list);
+  real = REAL_VALUE (value);
+
+  g_return_if_fail (real->d.list_data.type != GCONF_VALUE_INVALID);
+  g_return_if_fail ((list == NULL) ||
+                    ((list->data != NULL) &&
+                     (((GConfValue*)list->data)->type == real->d.list_data.type)));
+  
+  if (real->d.list_data.list)
+    gconf_value_free_list (value);
+
+  real->d.list_data.list = copy_value_list (list);
+}
+
+
+static int
+null_safe_strcmp (const char *lhs,
+                  const char *rhs)
+{
+  if (lhs == NULL && rhs == NULL)
+    return 0;
+  else if (lhs == NULL)
+    return -1;
+  else if (rhs == NULL)
+    return 1;
+  else
+    return strcmp (lhs, rhs);
+}
+
+int
+gconf_value_compare (const GConfValue *value_a,
+                     const GConfValue *value_b)
+{
+  g_return_val_if_fail (value_a != NULL, 0);
+  g_return_val_if_fail (value_b != NULL, 0);
+
+  /* Impose arbitrary type ordering, just to keep the
+   * sort invariants stable.
+   */
+  if (value_a->type < value_b->type)
+    return -1;
+  else if (value_a->type > value_b->type)
+    return 1;
+  
+  switch (value_a->type)
+    {
+    case GCONF_VALUE_INT:
+      if (gconf_value_get_int (value_a) < gconf_value_get_int (value_b))
+        return -1;
+      else if (gconf_value_get_int (value_a) > gconf_value_get_int (value_b))
+        return 1;
+      else
+        return 0;
+    case GCONF_VALUE_FLOAT:
+      if (gconf_value_get_float (value_a) < gconf_value_get_float (value_b))
+        return -1;
+      else if (gconf_value_get_float (value_a) > gconf_value_get_float (value_b))
+        return 1;
+      else
+        return 0;
+    case GCONF_VALUE_STRING:
+      return strcmp (gconf_value_get_string (value_a),
+                     gconf_value_get_string (value_b));
+    case GCONF_VALUE_BOOL:
+      if (gconf_value_get_bool (value_a) == gconf_value_get_bool (value_b))
+        return 0;
+      /* make TRUE > FALSE to maintain sort invariants */
+      else if (gconf_value_get_bool (value_a))
+        return 1;
+      else
+        return -1;
+    case GCONF_VALUE_LIST:
+      {
+        GSList *list_a;
+        GSList *list_b;
+
+        list_a = gconf_value_get_list (value_a);
+        list_b = gconf_value_get_list (value_b);
+        
+        while (list_a != NULL && list_b != NULL)
+          {
+            int result;
+
+            result = gconf_value_compare (list_a->data, list_b->data);
+
+            if (result != 0)
+              return result;
+            
+            list_a = g_slist_next (list_a);
+            list_b = g_slist_next (list_b);
+          }
+        
+        if (list_a)
+          return 1; /* list_a is longer so "greater" */
+        else if (list_b)
+          return -1;
+        else
+          return 0;
+      }
+    case GCONF_VALUE_PAIR:
+      {
+        GConfValue *a_car, *b_car, *a_cdr, *b_cdr;
+        int result;
+        
+        a_car = gconf_value_get_car (value_a);
+        b_car = gconf_value_get_car (value_b);
+        a_cdr = gconf_value_get_cdr (value_a);
+        b_cdr = gconf_value_get_cdr (value_b);
+
+        if (a_car == NULL && b_car != NULL)
+          return -1;
+        else if (a_car != NULL && b_car == NULL)
+          return 1;
+        else if (a_car != NULL && b_car != NULL)
+          {
+            result = gconf_value_compare (a_car, b_car);
+
+            if (result != 0)
+              return result;
+          }
+
+        if (a_cdr == NULL && b_cdr != NULL)
+          return -1;
+        else if (a_cdr != NULL && b_cdr == NULL)
+          return 1;
+        else if (a_cdr != NULL && b_cdr != NULL)
+          {
+            result = gconf_value_compare (a_cdr, b_cdr);
+
+            if (result != 0)
+              return result;
+          }
+
+        return 0;
+      }
+    case GCONF_VALUE_INVALID:
+      return 0;
+    case GCONF_VALUE_SCHEMA:
+      {
+        const char *locale_a, *locale_b;
+        GConfValueType type_a, type_b;
+        GConfValueType list_type_a, list_type_b;
+        GConfValueType car_type_a, car_type_b;
+        GConfValueType cdr_type_a, cdr_type_b;
+        const char *short_desc_a, *short_desc_b;
+        const char *long_desc_a, *long_desc_b;
+        int result;
+        
+        type_a = gconf_schema_get_type (gconf_value_get_schema (value_a));
+        type_b = gconf_schema_get_type (gconf_value_get_schema (value_b));
+
+        if (type_a < type_b)
+          return -1;
+        else if (type_a > type_b)
+          return 1;
+
+        short_desc_a = gconf_schema_get_short_desc (gconf_value_get_schema (value_a));
+        short_desc_b = gconf_schema_get_short_desc (gconf_value_get_schema (value_b));
+
+        result = null_safe_strcmp (short_desc_a, short_desc_b);
+        if (result != 0)
+          return result;
+        
+        long_desc_a = gconf_schema_get_long_desc (gconf_value_get_schema (value_a));
+
+
+        long_desc_b = gconf_schema_get_long_desc (gconf_value_get_schema (value_b));
+
+        result = null_safe_strcmp (long_desc_a, long_desc_b);
+        if (result != 0)
+          return result;
+        
+        locale_a = gconf_schema_get_locale (gconf_value_get_schema (value_a));
+        locale_b = gconf_schema_get_locale (gconf_value_get_schema (value_b));
+
+        result = null_safe_strcmp (locale_a, locale_b);
+        if (result != 0)
+          return result;        
+
+        if (type_a == GCONF_VALUE_LIST)
+          {
+            list_type_a = gconf_schema_get_list_type (gconf_value_get_schema (value_a));
+            list_type_b = gconf_schema_get_list_type (gconf_value_get_schema (value_b));
+            
+            if (list_type_a < list_type_b)
+              return -1;
+            else if (list_type_a > list_type_b)
+              return 1;
+          }
+
+        if (type_a == GCONF_VALUE_PAIR)
+          {
+            car_type_a = gconf_schema_get_car_type (gconf_value_get_schema (value_a));
+            car_type_b = gconf_schema_get_car_type (gconf_value_get_schema (value_b));
+            
+            if (car_type_a < car_type_b)
+              return -1;
+            else if (car_type_a > car_type_b)
+              return 1;
+            
+            cdr_type_a = gconf_schema_get_cdr_type (gconf_value_get_schema (value_a));
+            cdr_type_b = gconf_schema_get_cdr_type (gconf_value_get_schema (value_b));
+            
+            if (cdr_type_a < cdr_type_b)
+              return -1;
+            else if (cdr_type_a > cdr_type_b)
+              return 1;
+          }
+
+        return 0;
+      }
+    }
+
+  g_assert_not_reached ();
+
+  return 0;
 }
 
 /*
@@ -954,9 +1373,35 @@ gconf_meta_info_new(void)
 void
 gconf_meta_info_free(GConfMetaInfo* gcmi)
 {
+  set_string(&gcmi->schema, NULL);
+  set_string(&gcmi->mod_user, NULL);
   g_free(gcmi);
 }
 
+const char*
+gconf_meta_info_get_schema (GConfMetaInfo *gcmi)
+{
+  g_return_val_if_fail (gcmi != NULL, NULL);
+
+  return gcmi->schema;
+}
+
+const char*
+gconf_meta_info_get_mod_user (GConfMetaInfo *gcmi)
+{
+  g_return_val_if_fail (gcmi != NULL, NULL);
+
+  return gcmi->mod_user;
+}
+
+GTime
+gconf_meta_info_mod_time (GConfMetaInfo *gcmi)
+{
+  g_return_val_if_fail (gcmi != NULL, 0);
+  
+  return gcmi->mod_time;
+}
+
 void
 gconf_meta_info_set_schema  (GConfMetaInfo* gcmi,
                               const gchar* schema_name)
@@ -982,9 +1427,20 @@ gconf_meta_info_set_mod_time(GConfMetaInfo* gcmi,
  * GConfEntry
  */
 
+typedef struct {
+  char *key;
+  GConfValue *value;
+  char *schema_name;
+  int refcount;
+  guint is_default : 1;
+  guint is_writable : 1;
+} GConfRealEntry;
+
+#define REAL_ENTRY(x) ((GConfRealEntry*)(x))
+
 GConfEntry*
-gconf_entry_new (const gchar *key,
-                 GConfValue  *val)
+gconf_entry_new (const char *key,
+                 const GConfValue  *val)
 {
   return gconf_entry_new_nocopy (g_strdup (key),
                                  val ? gconf_value_copy (val) : NULL);
@@ -992,43 +1448,165 @@ gconf_entry_new (const gchar *key,
 }
 
 GConfEntry* 
-gconf_entry_new_nocopy(gchar* key, GConfValue* val)
+gconf_entry_new_nocopy (char* key, GConfValue* val)
 {
-  GConfEntry* pair;
+  GConfRealEntry* real;
 
-  pair = g_new(GConfEntry, 1);
+  real = g_slice_new (GConfRealEntry);
 
-  pair->key   = key;
-  pair->value = val;
-  pair->schema_name = NULL;
-  pair->is_default = FALSE;
-  pair->is_writable = TRUE;
+  real->key   = key;
+  real->value = val;
+  real->schema_name = NULL;
+  real->is_default = FALSE;
+  real->is_writable = TRUE;
+  real->refcount = 1;
   
-  return pair;
+  return (GConfEntry*) real;
 }
 
 void
-gconf_entry_free(GConfEntry* pair)
+gconf_entry_ref (GConfEntry *entry)
 {
-  g_free(pair->key);
-  if (pair->value)
-    gconf_value_free(pair->value);
-  if (pair->schema_name)
-    g_free (pair->schema_name);
-  g_free(pair);
+  g_return_if_fail (entry != NULL);
+  
+  REAL_ENTRY (entry)->refcount += 1;
+}
+
+void
+gconf_entry_unref (GConfEntry *entry)
+{
+  GConfRealEntry *real;
+  
+  g_return_if_fail (entry != NULL);
+  g_return_if_fail (REAL_ENTRY (entry)->refcount > 0);
+
+  real = REAL_ENTRY (entry);
+  
+  real->refcount -= 1;
+
+  if (real->refcount == 0)
+    {
+      g_free (real->key);
+      if (real->value)
+        gconf_value_free (real->value);
+      g_free (real->schema_name);
+      g_slice_free (GConfRealEntry, real);
+    }
+}
+
+void
+gconf_entry_free (GConfEntry *entry)
+{
+  gconf_entry_unref (entry);
+}
+
+GConfEntry*
+gconf_entry_copy (const GConfEntry *src)
+{
+  GConfEntry *entry;
+  GConfRealEntry *real;
+  
+  entry = gconf_entry_new (REAL_ENTRY (src)->key,
+                           REAL_ENTRY (src)->value);
+  real = REAL_ENTRY (entry);
+  
+  real->schema_name = g_strdup (REAL_ENTRY (src)->schema_name);
+  real->is_default = REAL_ENTRY (src)->is_default;
+  real->is_writable = REAL_ENTRY (src)->is_writable;
+
+  return entry;
+}
+
+gboolean
+gconf_entry_equal (const GConfEntry *a,
+                   const GConfEntry *b)
+{
+  GConfRealEntry *real_a;
+  GConfRealEntry *real_b;
+  
+  g_return_val_if_fail (a != NULL, FALSE);
+  g_return_val_if_fail (b != NULL, FALSE);
+
+  real_a = REAL_ENTRY (a);
+  real_b = REAL_ENTRY (b);
+  
+  /* do the cheap checks first, why not */
+  if (real_a->value && !real_b->value)
+    return FALSE;
+  else if (!real_a->value && real_b->value)
+    return FALSE;
+  else if (real_a->is_default != real_b->is_default)
+    return FALSE;
+  else if (real_a->is_writable != real_b->is_writable)
+    return FALSE;
+  else if (strcmp (real_a->key, real_b->key) != 0)
+    return FALSE;
+  else if (real_a->schema_name && !real_b->schema_name)
+    return FALSE;
+  else if (!real_a->schema_name && real_b->schema_name)
+    return FALSE;
+  else if (real_a->schema_name && real_b->schema_name &&
+           strcmp (real_a->schema_name, real_b->schema_name) != 0)
+    return FALSE;
+  else if (real_a->value && real_b->value &&
+           gconf_value_compare (real_a->value, real_b->value) != 0)
+    return FALSE;
+  else
+    return TRUE;
 }
 
 GConfValue*
 gconf_entry_steal_value (GConfEntry* entry)
 {
-  GConfValue* val = entry->value;
-  entry->value = NULL;
+  GConfValue* val = REAL_ENTRY (entry)->value;
+  REAL_ENTRY (entry)->value = NULL;
   return val;
 }
 
+const char*
+gconf_entry_get_key (const GConfEntry *entry)
+{
+  g_return_val_if_fail (entry != NULL, NULL);
+
+  return REAL_ENTRY (entry)->key;
+}
+
+GConfValue*
+gconf_entry_get_value (const GConfEntry *entry)
+{
+  g_return_val_if_fail (entry != NULL, NULL);
+
+  return REAL_ENTRY (entry)->value;
+}
+
+const char*
+gconf_entry_get_schema_name (const GConfEntry *entry)
+{
+  g_return_val_if_fail (entry != NULL, NULL);
+
+  return REAL_ENTRY (entry)->schema_name;
+}
+
+gboolean
+gconf_entry_get_is_default  (const GConfEntry *entry)
+{
+  g_return_val_if_fail (entry != NULL, FALSE);
+
+  return REAL_ENTRY (entry)->is_default;
+}
+
+gboolean
+gconf_entry_get_is_writable (const GConfEntry *entry)
+{
+  g_return_val_if_fail (entry != NULL, FALSE);
+
+  return REAL_ENTRY (entry)->is_writable;
+}
+
+
 void
 gconf_entry_set_value (GConfEntry  *entry,
-                       GConfValue  *val)
+                       const GConfValue  *val)
 {
   gconf_entry_set_value_nocopy (entry,
                                 val ? gconf_value_copy (val) : NULL);
@@ -1038,46 +1616,51 @@ void
 gconf_entry_set_value_nocopy(GConfEntry* entry,
                              GConfValue* val)
 {
-  if (entry->value)
-    gconf_value_free(entry->value);
+  if (REAL_ENTRY (entry)->value)
+    gconf_value_free (REAL_ENTRY (entry)->value);
 
-  entry->value = val;
+  REAL_ENTRY (entry)->value = val;
 }
 
 void
 gconf_entry_set_schema_name(GConfEntry* entry,
                             const gchar* name)
 {
-  if (entry->schema_name)
-    g_free(entry->schema_name);
+  g_free (REAL_ENTRY (entry)->schema_name);
 
-  entry->schema_name = name ? g_strdup(name) : NULL;
+  REAL_ENTRY (entry)->schema_name = g_strdup(name);
 }
 
 void
 gconf_entry_set_is_default (GConfEntry* entry,
                             gboolean is_default)
 {
-  entry->is_default = is_default;
+  REAL_ENTRY (entry)->is_default = is_default;
 }
 
 void
 gconf_entry_set_is_writable (GConfEntry  *entry,
                              gboolean     is_writable)
 {
-  entry->is_writable = is_writable;
+  REAL_ENTRY (entry)->is_writable = is_writable;
 }
 
 
 gboolean
-gconf_value_validate (GConfValue *value,
-                      GError    **err)
+gconf_value_validate (const GConfValue *value,
+                      GError          **err)
 {
+  GConfRealValue *real;
+
+  g_return_val_if_fail (value != NULL, FALSE);
+
+  real = REAL_VALUE (value);
+  
   switch (value->type)
     {
     case GCONF_VALUE_STRING:
-      if (value->d.string_data &&
-          !g_utf8_validate (value->d.string_data, -1, NULL))
+      if (real->d.string_data &&
+          !g_utf8_validate (real->d.string_data, -1, NULL))
         {
           g_set_error (err, GCONF_ERROR,
                        GCONF_ERROR_FAILED,
@@ -1087,8 +1670,8 @@ gconf_value_validate (GConfValue *value,
       break;
 
     case GCONF_VALUE_SCHEMA:
-      if (value->d.schema_data)
-        return gconf_schema_validate (value->d.schema_data,
+      if (real->d.schema_data)
+        return gconf_schema_validate (real->d.schema_data,
                                       err);
       break;
 
